// adu.c - command line utility to control an Ontrak ADU card
// Copyright 2005 by John Homppi
// Released to the public domain. You may use this program
// or portions of it in your own commercial or non-commercial
// applications without royalty payments, without attributions, 
// without restrictions.

// adu is invoked from the command line to control an ADU board.
// Usage:
//   adu [-i <0-9>] [-s serialnumber] [-r] [-f filename] [-v] [cmd]
// where:
//   -r read a response from the ADU board  
//   -i specifies the device id to append to the device name
//      ie 0 means /dev/usb/adutux0
//   -s specifies the serial number (in lieu of device id)
//   -f filename is the name of a file containing commands
//   -v verbose results are printed on stderr
//   cmd if present is sent to the ADU device before the optional read
//
// If none of the -r, -f, cmd options are present then standard
// input is opened for commands.
//
// If -s is specified then each ADU device name (adutux0 thru 
// adutux9) is opened and checked for a serial number match.
// The command is written to the matching device.
//
// If neither -i nor -s is specified then the action is performed
// against the first /dev/usb/adutuxN device that opens
// successfully. 
// 
// Format for file and console input
//   <w|r> [cmd]
// Each input line consists of a read/write indicator followed
// by a command.
// eg.
// w mk1
// w mk2
// w rpk
// r 
// w mk3

// is equivalent to
// w mk1
// w mk2
// r rpk
// w mk3

// Notes:
// All reads and writes are directed to the command pipe of the
// ADU board (ie pipe 1).
// Returns: 0 = success, non-zero = failure

#define VERSION "alpha 0.0.1"
#define DEBUG_ADU
#undef DEBUG_ADU
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <getopt.h>
#include <ctype.h>
#define MAXFILENAME 64
#define MAXSERIAL 6
#define MAXCOMMAND 7
// USB 1.1 low speed latency can be as much as 10 milli-seconds
#define USBLATENCY 10000

static int gDoRead;
static int gDoVersion;
static int gId;
static int gFd;
static FILE * gFdCmd;
static char gFilename[MAXFILENAME + 1];
static char gSerial[MAXSERIAL + 1];
static char gCommand[MAXCOMMAND + 1];

static const struct option long_options[] =
{
	{ "filename", required_argument, NULL, 'f' },
	{ "id", required_argument, NULL, 'i' },
	{ "read", no_argument, NULL, 'r'},
    { "debug", no_argument, NULL, 'd'},
    { "version", no_argument, NULL, 'v'},
	{ "serial", required_argument, NULL, 's' },
	{ NULL, 0, NULL, 0 }
};

void usage (void)
{
	fprintf(stderr, "usage: aducmd -i id -s serial -r -v -f filename command\n");
	exit (-1);
} // end usage

void parse_args (int argc, char *argv[])
{
	int opt;

	gFilename[0] = '\0';
	gSerial[0] = '\0';
	gCommand[0] = '\0';
	gId = -1;
	gDoRead = 0;
	gDoRead = 0;
  
	while ((opt = getopt_long (argc, argv, "f:i:rqvs:", long_options, NULL)) != -1)
	{
		switch (opt)
		{
			case 'f':
				strncpy(gFilename, optarg, MAXFILENAME);
				break;
			case 'i':
				if (!isdigit(*optarg)) 
				{
					fprintf(stderr, "bad id %c, must be a numeric digit\n", *optarg);
					exit(-1);
				}
				gId = atoi(optarg);
				break;
			case 'r':
				gDoRead = 1;;
				break;
			case 'v':
				gDoVersion = 1;;
				break;
			case 's':
				strncpy(gSerial, optarg, MAXSERIAL);
				break;
			case 'h':
			default:
				usage ();
		} // end switch
	} // end while
	if (argc > optind)
	{
		strncpy(gCommand, argv[optind], MAXCOMMAND);
	} // end if
#ifdef DEBUG_ADU
	printf("filename is %s\n", gFilename);
	printf("serial is %s\n", gSerial);
	printf("id is %d\n", gId);
	printf("do_read is %d\n", gDoRead);
	printf("command is %s\n", gCommand);
#endif
} // end parse_args

void writeCommand (char *sCommand)
{
	int status;
	char sString[8];
	memset(sString, 0, 8);
	sprintf(sString, "%c%s", 1, sCommand);
	status = write(gFd, sString, 8);
	usleep(USBLATENCY);  // give the ADU time to receive the command
					
	if (status < 0)
	{
		fprintf(stderr, "write failed. status=%d errno=%d %s\n", 
				status, errno, strerror(errno));
		exit (errno);
	} // end if
} // end writeCommand

 void readResponse (void)
 {
	int status;
	char sString[8];
	char sResponse[65];
	
	memset(sString, 0, 8);
	status = read(gFd, sString, 8);
	usleep(USBLATENCY);  // give the ADU some time
	memset(sResponse, 0, 9);
	strncpy(sResponse, sString, 8);
	if (status < 0)
	{
		fprintf(stderr, "\nread failed. errno=%d %s\n", errno, strerror(errno));
		exit (errno);
	}
	else
	{
		fprintf(stdout, "%s\n", &sResponse[1]);
#ifdef DEBUG_ADU
		printf("response=%02x%02x%02x%02x%02x%02x%02x%02x %s\n",
			sResponse[0], sResponse[1], sResponse[2], sResponse[3], 
			sResponse[4], sResponse[5], sResponse[6], sResponse[7],
			sResponse);
#endif
	} // end if
} // end readResponse

void doCommands (void)
{
	char sInput[256];
	int  status;
	char cAction;
	char sCommand[256];
	char sResponse[65];

	for (;1;)
	{
		{
			int i;
			char cChar;
			char * pIn;
			memset(sInput, 0, 256);
			pIn = sInput;
			for (i=0 ; i<256 ; i++)
			{
				cChar = fgetc(gFdCmd);
				if (cChar == EOF) return;
				*pIn = cChar;
				pIn++;
				if (cChar == '\n') break;
			}
			pIn = '\0';
		}
#ifdef DEBUG_ADU
		printf("read status is %d\n", status);
		printf("string that was read is %s\n", sInput); 
#endif
		cAction = '0';
		memset(sCommand, 0 , 256);
		
		sscanf(sInput, "%c %s", &cAction, sCommand);
		switch(cAction)
		{
			case('w'):
			{
				writeCommand(sCommand);
				break;
			} // end case 'w'
			case('r'):
			{
				if (strlen(sCommand) > 0)
				{
					writeCommand(sCommand);
				} // end if
				readResponse();
				break;
			} // end case 'r'
			case('s'):
			{
				char sString[8];
				memset(sString, 0, 8);
				sString[0] = 's';
				status = read(gFd, sString, 8);
				// no need for usleep - serial number is cached in driver
				memset(sResponse, 0, 9);
				strncpy(sResponse, sString, 8);
				if (status < 0)
				{
					fprintf(stderr, "\nread failed. status=%d errno=%d %s\n", 
							status, errno, strerror(errno));
					exit (errno);
				}
				else
				{
					fprintf(stdout, "%s\n",sResponse);
#ifdef DEBUG_ADU
					printf("serial number=%02x%02x%02x%02x%02x%02x%02x%02x %s\n",
						sResponse[0], sResponse[1], sResponse[2], sResponse[3], 
						sResponse[4], sResponse[5], sResponse[6], sResponse[7],
						sResponse);
#endif
				} // end if
				break;
			} // end case 's'
			default:
			{
				fprintf(stderr, "invalid action '%c' ignored\n", cAction);
			} // end default
		} // end switch
	} // end for
} // end doCommands

int main (int argc, char *argv[])
{
	int id;
	int status;
	char sFilename[256];
	gFd = 0;
	
	parse_args(argc, argv);
	
	if (gDoVersion)
	{
		fprintf(stdout, "adu command line utility **EXPERIMENTAL** "
			"Version %s\nCompiled %s %s\n"
			"Copyright 2005 by John Homppi\n", 
			VERSION, __DATE__, __TIME__);
	} // end if
	
	if (gId > -1) // specific device name was requested
	{
		sprintf(sFilename, "/dev/usb/adutux%d", gId);
		gFd = open(sFilename, O_RDWR);
	
    	if (gFd < 0)
    	{
   		fprintf(stderr, "Open failed on %s. errno=%d %s\n", sFilename, errno, 
   				strerror(errno));
   		return -1;
    	}
    	else
    	{
#ifdef DEBUG_ADU
   		printf("Opened %s\n", sFilename);
#endif
    	} // end if
    }
    else // user did not request a specific device so must search for one 
    {
    	for (id = 0; id < 10; id++)
    	{
    		sprintf(sFilename, "/dev/usb/adutux%d", id);
			gFd = open(sFilename, O_RDWR);
			if (gFd > 0) // there is an ADU attached to this device name
			{
				if (strlen(gSerial) > 0) // user wants specific serial number
				{
					char sString[8];
					memset(sString, 0, 8);
					sString[0] = 's'; // get serial number from adutux driver
					status = read(gFd, sString, 8);
					// no need for usleep, serial number is cached in driver
					if (status < 1 || strcmp(sString, gSerial))
					{
						// not this device - continue searching
						close(gFd);
						gFd = 0;
					}
					else
					{	
						break; // serial matches so use this device
					} // end if
				}
				else
				{
					break; // no serial number specified so use first device found				
				} // end if
			} // end if
    	} // end for
    } // end if
   		
	if (gFd < 1)
	{
		fprintf(stderr, "no adutux device found\n");
		exit (-1);
	} // end if
	
	if (strlen(gCommand) > 0) // was the optional command arg specified
	{
		writeCommand(gCommand);
		usleep(10000);  // give the ADU time to receive the command
						// USB 1.1 low speed latency can be 10 milli-seconds
	} 
	else if (strlen(gFilename) > 0) // was a filename specified for input
	{
		gFdCmd = fopen(gFilename, "r");
		if (gFdCmd > 0)
		{
			doCommands();
		} // end if
	}
	else if (gDoRead == 0) // was not a read-only request 
	{
		gFdCmd = stdin; // default is stdin (console or pipe)
		doCommands();
	} // end if
	
	if (gDoRead > 0) // did the user request a read operation 
	{
		readResponse();
	} // end if
	
	if (gFd > 0)
	{
		close(gFd);
		gFd = 0;
	} // end if
	return 0;
} // end main

// end aducmd.c
